import contextvars

from typing import List, Optional
from opik.api_objects import span, trace

_current_trace_data_context: contextvars.ContextVar[Optional[trace.TraceData]] = (
    contextvars.ContextVar("current_trace_data", default=None)
)

_spans_data_stack_context: contextvars.ContextVar[List[span.SpanData]] = (
    contextvars.ContextVar("spans_data_stack", default=[])
)

# Read if you are going to change this module.
#
# If you want to add or pop the span from the stack
# you must NOT do it like that: spans_stack_context.get().append()
#
# If you modify the default object in ContextVar (which is a MUTABLE list)
# these changes will be reflected in any new context which breaks the whole idea
# of ContextVar.
#
# You should append to it like that:
# span_data_stack = span_data_stack_context.get().copy()
# span_data_stack_context.set(span_data_stack + [new_element])
#
# And pop like that:
# span_data_stack = spans_stack_context.get().copy()
# span_data_stack_context.set(span_data_stack[:-1])
#
# The following functions provide an API to work with ContextVars this way


def top_span_data() -> Optional[span.SpanData]:
    if span_data_stack_empty():
        return None

    stack = _get_data_stack()
    return stack[-1]


def pop_span_data() -> Optional[span.SpanData]:
    stack = _get_data_stack()
    _spans_data_stack_context.set(stack[:-1])
    return stack[-1]


def add_span_data(span: span.SpanData) -> None:
    stack = _get_data_stack()
    _spans_data_stack_context.set(stack + [span])


def span_data_stack_empty() -> bool:
    return len(_get_data_stack()) == 0


def _get_data_stack() -> List[span.SpanData]:
    return _spans_data_stack_context.get().copy()


def get_trace_data() -> Optional[trace.TraceData]:
    trace = _current_trace_data_context.get()
    return trace


def pop_trace_data() -> Optional[trace.TraceData]:
    trace = _current_trace_data_context.get()
    set_trace_data(None)
    return trace


def set_trace_data(trace: Optional[trace.TraceData]) -> None:
    _current_trace_data_context.set(trace)


def clear_all() -> None:
    _current_trace_data_context.set(None)
    _spans_data_stack_context.set([])



import logging
import json
from typing import List, Any, Dict, Optional, Callable, Tuple, Union
from opik.types import SpanType
from opik.decorator import base_track_decorator, arguments_helpers
from . import stream_wrappers, chunks_aggregator

import openai
from openai.types.chat import chat_completion, chat_completion_message

LOGGER = logging.getLogger(__name__)

CreateCallResult = Union[chat_completion.ChatCompletion, List[Any]]


class OpenaiTrackDecorator(base_track_decorator.BaseTrackDecorator):
    """
    An implementation of BaseTrackDecorator designed specifically for tracking
    calls of OpenAI().chat.completion.create function.

    Besides special processing for input arguments and response content, it
    overrides _generators_handler() method to work correctly with
    openai.Stream and openai.AsyncStream objects.
    """

    def _start_span_inputs_preprocessor(
        self,
        func: Callable,
        name: Optional[str],
        type: SpanType,
        tags: Optional[List[str]],
        metadata: Optional[Dict[str, Any]],
        capture_input: bool,
        args: Optional[Tuple],
        kwargs: Optional[Dict[str, Any]],
        project_name: Optional[str],
    ) -> arguments_helpers.StartSpanParameters:
        assert (
            kwargs is not None
        ), "Expected kwargs to be not None in OpenAI().chat.completion.create(**kwargs)"
        kwargs_copy = kwargs.copy()

        name = name if name is not None else func.__name__

        input = {}
        input["messages"] = _parse_messages_list(kwargs_copy.pop("messages"))
        if "function_call" in kwargs_copy:
            input["function_call"] = kwargs_copy.pop("function_call")

        metadata = {
            "created_from": "openai",
            "type": "openai_chat",
            **kwargs_copy,
        }

        tags = ["openai"]

        result = arguments_helpers.StartSpanParameters(
            name=name,
            input=input,
            type=type,
            tags=tags,
            metadata=metadata,
            project_name=project_name,
        )

        return result

    def _end_span_inputs_preprocessor(
        self, output: Any, capture_output: bool
    ) -> arguments_helpers.EndSpanParameters:
        assert isinstance(
            output,
            (chat_completion.ChatCompletion, chunks_aggregator.ExtractedStreamContent),
        )

        usage = None

        if isinstance(output, chat_completion.ChatCompletion):
            result_dict = output.model_dump(mode="json")
            choices: List[Dict[str, Any]] = result_dict.pop("choices")  # type: ignore
            output = {"choices": choices}

            usage = result_dict["usage"]
        elif isinstance(output, chunks_aggregator.ExtractedStreamContent):
            usage = output.usage
            output = {"choices": output.choices}

        result = arguments_helpers.EndSpanParameters(output=output, usage=usage)

        return result

    def _generators_handler(
        self,
        output: Any,
        capture_output: bool,
        generations_aggregator: Optional[Callable[[List[Any]], Any]],
    ) -> Union[
        base_track_decorator.Generator[Any, None, None],
        base_track_decorator.AsyncGenerator[Any, None],
        None,
    ]:
        assert (
            generations_aggregator is not None
        ), "OpenAI decorator will always get aggregator function as input"

        if isinstance(output, openai.Stream):
            span_to_end, trace_to_end = base_track_decorator.pop_end_candidates()
            return stream_wrappers.wrap_sync_stream(
                generator=output,
                capture_output=capture_output,
                span_to_end=span_to_end,
                trace_to_end=trace_to_end,
                generations_aggregator=generations_aggregator,
                finally_callback=self._after_call,
            )

        if isinstance(output, openai.AsyncStream):
            span_to_end, trace_to_end = base_track_decorator.pop_end_candidates()
            return stream_wrappers.wrap_async_stream(
                generator=output,
                capture_output=capture_output,
                span_to_end=span_to_end,
                trace_to_end=trace_to_end,
                generations_aggregator=generations_aggregator,
                finally_callback=self._after_call,
            )

        NOT_A_STREAM = None

        return NOT_A_STREAM


def _parse_messages_list(
    messages: List[
        Union[Dict[str, Any], chat_completion_message.ChatCompletionMessage]
    ],
) -> List[Dict[str, Any]]:
    if _is_jsonable(messages):
        return messages

    result = []

    for message in messages:
        if _is_jsonable(message):
            result.append(message)
            continue

        if isinstance(message, chat_completion_message.ChatCompletionMessage):
            result.append(message.model_dump(mode="json"))
            continue

        LOGGER.debug("Message %s is not json serializable", message)

        result.append(
            str(message)
        )  # TODO: replace with Opik serializer when it is implemented

    return result


def _is_jsonable(x: Any) -> bool:
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False



from typing import Optional, Dict, Any
from llama_index.core.callbacks import schema as llama_index_schema


def get_span_input_from_events(
    event_type: llama_index_schema.CBEventType, payload: Optional[Dict[str, Any]]
) -> Optional[Dict[str, Any]]:
    if payload is None:
        return None

    payload_copy = payload.copy()

    if llama_index_schema.EventPayload.SERIALIZED in payload:
        # Remove
        payload_copy.pop(llama_index_schema.EventPayload.SERIALIZED)

    if (
        event_type == llama_index_schema.CBEventType.EMBEDDING
        and llama_index_schema.EventPayload.CHUNKS in payload_copy
    ):
        chunks = payload_copy.get(llama_index_schema.EventPayload.CHUNKS, [])
        return {"number_of_chunks": len(chunks)}

    if (
        event_type == llama_index_schema.CBEventType.NODE_PARSING
        and llama_index_schema.EventPayload.DOCUMENTS in payload_copy
    ):
        documents = payload_copy.pop(llama_index_schema.EventPayload.DOCUMENTS)
        payload_copy["documents"] = [doc.metadata for doc in documents]
        return payload_copy

    if llama_index_schema.EventPayload.QUERY_STR in payload_copy:
        return {
            llama_index_schema.EventPayload.QUERY_STR: payload_copy.get(
                llama_index_schema.EventPayload.QUERY_STR
            )
        }

    if llama_index_schema.EventPayload.MESSAGES in payload_copy:
        return {"messages": payload_copy.get(llama_index_schema.EventPayload.MESSAGES)}

    if llama_index_schema.EventPayload.PROMPT in payload_copy:
        return {"prompt": payload_copy.get(llama_index_schema.EventPayload.PROMPT)}

    if payload_copy:
        return {"input": payload_copy}
    else:
        return None


def get_span_output_from_event(
    event_type: llama_index_schema.CBEventType, payload: Optional[Dict[str, Any]]
) -> Optional[Dict[str, Any]]:
    if payload is None:
        return None

    payload_copy = payload.copy()

    if llama_index_schema.EventPayload.SERIALIZED in payload_copy:
        # Always pop Serialized from payload as it may contain LLM api keys
        payload_copy.pop(llama_index_schema.EventPayload.SERIALIZED)

    if (
        event_type == llama_index_schema.CBEventType.EMBEDDING
        and llama_index_schema.EventPayload.EMBEDDINGS in payload_copy
    ):
        embeddings: Any = payload_copy.get(llama_index_schema.EventPayload.EMBEDDINGS)
        return {"num_embeddings": len(embeddings)}

    if (
        event_type == llama_index_schema.CBEventType.NODE_PARSING
        and llama_index_schema.EventPayload.NODES in payload_copy
    ):
        nodes = payload_copy.pop(llama_index_schema.EventPayload.NODES)
        payload_copy["num_nodes"] = len(nodes)
        return {"output": payload_copy}

    if (
        event_type == llama_index_schema.CBEventType.CHUNKING
        and llama_index_schema.EventPayload.CHUNKS in payload
    ):
        chunks = payload.pop(llama_index_schema.EventPayload.CHUNKS)
        payload_copy["num_chunks"] = len(chunks)

    if llama_index_schema.EventPayload.COMPLETION in payload:
        return {"output": payload_copy.get(llama_index_schema.EventPayload.COMPLETION)}

    if llama_index_schema.EventPayload.RESPONSE in payload:
        response: Any = payload_copy.get(llama_index_schema.EventPayload.RESPONSE)

        # Skip streaming responses as consuming them would block the user's execution path
        if "Streaming" in type(response).__name__:
            return None

        if hasattr(response, "response"):
            return {"output": response.response}

        if hasattr(response, "message"):
            output = dict(response.message)
            if "additional_kwargs" in output:
                if "tool_calls" in output["additional_kwargs"]:
                    output["tool_calls"] = output["additional_kwargs"]["tool_calls"]

                del output["additional_kwargs"]

            return {"output": output}

    if payload_copy:
        return {"output": payload_copy}
    else:
        return None



from typing import Callable, Any, Dict, Tuple
import os
import functools
import logging
import opik
from opik import opik_context
from . import test_runs_storage, test_run_content
from opik.decorator import inspect_helpers
from opik import config

LOGGER = logging.getLogger(__name__)


def llm_unit(
    expected_output_key: str = "expected_output",
    input_key: str = "input",
    metadata_key: str = "metadata",
) -> Callable[[Any], Any]:
    """
    Decorator used for special tests tracking.
    Mark your test with `llm_unit` and when you run `pytest`, Opik will
    create an experiment and log test results to it: test name, test inputs, result.

    Arguments:
        expected_output_key: test argument name that will be logged as `expected_output` of the LLM task.
            If not provided, Opik will try to find `expected_output` in arguments.
        input_key: test argument name that will be logged as `input` of the LLM task.
            If not provided, Opik will try to find `input` in arguments.
        metadata_key: test argument name that will be logged as `metadata`.
            If not provided, Opik will try to find `metadata` in arguments.
    """
    argnames_mapping = {
        "expected_output": expected_output_key,
        "input": input_key,
        "metadata": metadata_key,
    }

    def decorator(func: Callable[[Any], Any]) -> Callable[[Any], Any]:
        config_ = config.get_from_user_inputs()
        if not config_.pytest_experiment_enabled:
            return func

        @opik.track(capture_input=False)
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                test_trace_data = opik_context.get_current_trace_data()
                test_span_data = opik_context.get_current_span_data()
                assert (
                    test_trace_data is not None and test_span_data is not None
                ), "Must not be None here by design assumption"

                node_id: str = _get_test_nodeid()
                test_runs_storage.LLM_UNIT_TEST_RUNS.add(node_id)

                test_run_content_ = _get_test_run_content(
                    func=func,
                    args=args,
                    kwargs=kwargs,
                    argnames_mapping=argnames_mapping,
                )

                trace_input = {**test_run_content_.input}
                trace_input.pop("test_name")  # we don't need it in traces
                opik_context.update_current_trace(
                    input=trace_input,
                    metadata=test_run_content_.metadata,
                )
                opik_context.update_current_span(
                    input=trace_input,
                    metadata=test_run_content_.metadata,
                )

                test_runs_storage.TEST_RUNS_TO_TRACE_DATA[node_id] = test_trace_data
                test_runs_storage.TEST_RUNS_CONTENTS[node_id] = test_run_content_
            except Exception:
                LOGGER.error(
                    "Unexpected exception occured during llm_unit test tracking for test %s",
                    func.__name__,
                    exc_info=True,
                )

            result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator


def _get_test_nodeid() -> str:
    # Examples of environment variables:
    # 'sdks/python/tests/tests_sandbox/test_things.py::TestGroup::test_example[13 32] (call)'
    # 'sdks/python/tests/tests_sandbox/test_things.py::TestGroup::test_example (call)'
    # 'sdks/python/tests/tests_sandbox/test_things.py::test_example (call)'

    return os.environ["PYTEST_CURRENT_TEST"].rpartition(" ")[0]


def _get_test_run_content(
    func: Callable,
    args: Tuple,
    kwargs: Dict[str, Any],
    argnames_mapping: Dict[str, str],
) -> test_run_content.TestRunContent:
    test_inputs = inspect_helpers.extract_inputs(func, args, kwargs)
    input = test_inputs.get(argnames_mapping["input"], {})
    metadata = test_inputs.get(argnames_mapping["metadata"], None)
    expected_output = test_inputs.get(argnames_mapping["expected_output"], None)

    if not isinstance(input, dict):
        input = {"test_name": _get_test_nodeid(), "input": input}
    else:
        input = {"test_name": _get_test_nodeid(), **input}

    if expected_output is not None and not isinstance(expected_output, dict):
        expected_output = {"expected_output": expected_output}

    if metadata is not None and not isinstance(metadata, dict):
        metadata = {"metadata": metadata}

    result = test_run_content.TestRunContent(
        input=input,
        expected_output=expected_output,
        metadata=metadata,
    )

    return result



import functools
import logging
import inspect
import abc
from typing import (
    List,
    Any,
    Dict,
    Set,
    Optional,
    Callable,
    Tuple,
    Generator,
    Union,
    AsyncGenerator,
)

from ..types import SpanType, DistributedTraceHeadersDict
from . import arguments_helpers, generator_wrappers, inspect_helpers
from ..api_objects import opik_client, helpers, span, trace
from .. import context_storage, logging_messages, datetime_helpers

LOGGER = logging.getLogger(__name__)

TRACES_CREATED_BY_DECORATOR: Set[str] = set()


class BaseTrackDecorator(abc.ABC):
    """
    For internal usage.

    All TrackDecorator instances share the same context and can be
    used together simultaneously.

    The following methods MUST be implemented in the subclass:
        * _start_span_inputs_preprocessor
        * _end_span_inputs_preprocessor

    The following methods CAN be overriden in the subclass:
        * _generators_handler

    Overriding other methods of this class is not recommended.
    """

    def track(
        self,
        name: Optional[Union[Callable, str]] = None,
        type: SpanType = "general",
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        capture_input: bool = True,
        capture_output: bool = True,
        generations_aggregator: Optional[Callable[[List[Any]], Any]] = None,
        flush: bool = False,
        project_name: Optional[str] = None,
    ) -> Union[Callable, Callable[[Callable], Callable]]:
        """
        Decorator to track the execution of a function.

        Can be used as @track or @track().

        Args:
            name: The name of the span.
            type: The type of the span.
            tags: Tags to associate with the span.
            metadata: Metadata to associate with the span.
            capture_input: Whether to capture the input arguments.
            capture_output: Whether to capture the output result.
            generations_aggregator: Function to aggregate generation results.
            flush: Whether to flush the client after logging.
            project_name: The name of the project to log data.

        Returns:
            Callable: The decorated function(if used without parentheses)
                or the decorator function (if used with parentheses).

        Note:
            You can use this decorator to track nested functions, Opik will automatically create
            a trace and correctly span nested function calls.

            This decorator can be used to track both synchronous and asynchronous functions,
            and also synchronous and asynchronous generators.
            It automatically detects the function type and applies the appropriate tracking logic.
        """
        if callable(name):
            # Decorator was used without '()'. It means that decorated function
            # automatically passed as the first argument of 'track' function - name
            func = name
            return self._decorate(
                func=func,
                name=None,
                type=type,
                tags=tags,
                metadata=metadata,
                capture_input=capture_input,
                capture_output=capture_output,
                generations_aggregator=generations_aggregator,
                flush=flush,
                project_name=project_name,
            )

        def decorator(func: Callable) -> Callable:
            return self._decorate(
                func=func,
                name=name,
                type=type,
                tags=tags,
                metadata=metadata,
                capture_input=capture_input,
                capture_output=capture_output,
                generations_aggregator=generations_aggregator,
                flush=flush,
                project_name=project_name,
            )

        return decorator

    def _decorate(
        self,
        func: Callable,
        name: Optional[str],
        type: SpanType,
        tags: Optional[List[str]],
        metadata: Optional[Dict[str, Any]],
        capture_input: bool,
        capture_output: bool,
        generations_aggregator: Optional[Callable[[List[Any]], Any]],
        flush: bool,
        project_name: Optional[str],
    ) -> Callable:
        if not inspect_helpers.is_async(func):
            return self._tracked_sync(
                func=func,
                name=name,
                type=type,
                tags=tags,
                metadata=metadata,
                capture_input=capture_input,
                capture_output=capture_output,
                generations_aggregator=generations_aggregator,
                flush=flush,
                project_name=project_name,
            )

        return self._tracked_async(
            func=func,
            name=name,
            type=type,
            tags=tags,
            metadata=metadata,
            capture_input=capture_input,
            capture_output=capture_output,
            generations_aggregator=generations_aggregator,
            flush=flush,
            project_name=project_name,
        )

    def _tracked_sync(
        self,
        func: Callable,
        name: Optional[str],
        type: SpanType,
        tags: Optional[List[str]],
        metadata: Optional[Dict[str, Any]],
        capture_input: bool,
        capture_output: bool,
        generations_aggregator: Optional[Callable[[List[Any]], str]],
        flush: bool,
        project_name: Optional[str],
    ) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:  # type: ignore
            self._before_call(
                func=func,
                name=name,
                type=type,
                tags=tags,
                metadata=metadata,
                capture_input=capture_input,
                project_name=project_name,
                args=args,
                kwargs=kwargs,
            )

            result = None
            try:
                result = func(*args, **kwargs)
            except Exception as exception:
                LOGGER.debug(
                    logging_messages.EXCEPTION_RAISED_FROM_TRACKED_FUNCTION,
                    func.__name__,
                    (args, kwargs),
                    str(exception),
                    exc_info=True,
                )
                raise exception
            finally:
                generator_or_generator_container = self._generators_handler(
                    result,
                    capture_output,
                    generations_aggregator,
                )
                if generator_or_generator_container is not None:
                    return generator_or_generator_container

                self._after_call(
                    output=result,
                    capture_output=capture_output,
                    flush=flush,
                )
                if result is not None:
                    return result

        wrapper.opik_tracked = True  # type: ignore

        return wrapper

    def _tracked_async(
        self,
        func: Callable,
        name: Optional[str],
        type: SpanType,
        tags: Optional[List[str]],
        metadata: Optional[Dict[str, Any]],
        capture_input: bool,
        capture_output: bool,
        generations_aggregator: Optional[Callable[[List[Any]], str]],
        flush: bool,
        project_name: Optional[str],
    ) -> Callable:
        @functools.wraps(func)
        async def wrapper(*args, **kwargs) -> Any:  # type: ignore
            self._before_call(
                func=func,
                name=name,
                type=type,
                tags=tags,
                metadata=metadata,
                capture_input=capture_input,
                project_name=project_name,
                args=args,
                kwargs=kwargs,
            )
            result = None
            try:
                result = await func(*args, **kwargs)
            except Exception as exception:
                LOGGER.error(
                    logging_messages.EXCEPTION_RAISED_FROM_TRACKED_FUNCTION,
                    func.__name__,
                    (args, kwargs),
                    str(exception),
                    exc_info=True,
                )
                raise exception
            finally:
                generator = self._generators_handler(
                    result,
                    capture_output,
                    generations_aggregator,
                )
                if generator is not None:  # TODO: test this flow for async generators
                    return generator

                self._after_call(
                    output=result,
                    capture_output=capture_output,
                    flush=flush,
                )
                if result is not None:
                    return result

        wrapper.opik_tracked = True  # type: ignore
        return wrapper

    def _before_call(
        self,
        func: Callable,
        name: Optional[str],
        type: SpanType,
        tags: Optional[List[str]],
        metadata: Optional[Dict[str, Any]],
        capture_input: bool,
        project_name: Optional[str],
        args: Tuple,
        kwargs: Dict[str, Any],
    ) -> None:
        try:
            opik_distributed_trace_headers: Optional[DistributedTraceHeadersDict] = (
                kwargs.pop("opik_distributed_trace_headers", None)
            )

            start_span_arguments = self._start_span_inputs_preprocessor(
                func=func,
                name=name,
                type=type,
                tags=tags,
                metadata=metadata,
                capture_input=capture_input,
                project_name=project_name,
                args=args,
                kwargs=kwargs,
            )

            if opik_distributed_trace_headers is None:
                self._create_span(start_span_arguments)
            else:
                self._create_distributed_node_root_span(
                    start_span_arguments, opik_distributed_trace_headers
                )

        except Exception as exception:
            LOGGER.error(
                logging_messages.UNEXPECTED_EXCEPTION_ON_SPAN_CREATION_FOR_TRACKED_FUNCTION,
                func.__name__,
                (args, kwargs),
                str(exception),
                exc_info=True,
            )

    def _create_span(
        self, start_span_arguments: arguments_helpers.StartSpanParameters
    ) -> None:
        """
        Handles different span creation flows.
        """
        current_span_data = context_storage.top_span_data()
        current_trace_data = context_storage.get_trace_data()

        span_data: span.SpanData
        trace_data: trace.TraceData

        if current_span_data is not None:
            # There is already at least one span in current context.
            # Simply attach a new span to it.
            assert current_trace_data is not None

            project_name = helpers.resolve_child_span_project_name(
                parent_project_name=current_span_data.project_name,
                child_project_name=start_span_arguments.project_name,
                show_warning=current_trace_data.created_by != "evaluation",
            )

            start_span_arguments.project_name = project_name

            span_data = span.SpanData(
                id=helpers.generate_id(),
                parent_span_id=current_span_data.id,
                trace_id=current_span_data.trace_id,
                start_time=datetime_helpers.local_timestamp(),
                name=start_span_arguments.name,
                type=start_span_arguments.type,
                tags=start_span_arguments.tags,
                metadata=start_span_arguments.metadata,
                input=start_span_arguments.input,
                project_name=start_span_arguments.project_name,
            )
            context_storage.add_span_data(span_data)
            return

        if current_trace_data is not None and current_span_data is None:
            # By default, we expect trace to be created with a span.
            # But there can be cases when trace was created and added
            # to context manually (not via decorator).
            # In that case decorator should just create a span for the existing trace.

            project_name = helpers.resolve_child_span_project_name(
                parent_project_name=current_trace_data.project_name,
                child_project_name=start_span_arguments.project_name,
                show_warning=current_trace_data.created_by != "evaluation",
            )

            start_span_arguments.project_name = project_name

            span_data = span.SpanData(
                id=helpers.generate_id(),
                parent_span_id=None,
                trace_id=current_trace_data.id,
                start_time=datetime_helpers.local_timestamp(),
                name=start_span_arguments.name,
                type=start_span_arguments.type,
                tags=start_span_arguments.tags,
                metadata=start_span_arguments.metadata,
                input=start_span_arguments.input,
                project_name=start_span_arguments.project_name,
            )
            context_storage.add_span_data(span_data)
            return

        if current_span_data is None and current_trace_data is None:
            # Create a trace and root span because it is
            # the first decorated function run in current context.
            trace_data = trace.TraceData(
                id=helpers.generate_id(),
                start_time=datetime_helpers.local_timestamp(),
                name=start_span_arguments.name,
                input=start_span_arguments.input,
                metadata=start_span_arguments.metadata,
                tags=start_span_arguments.tags,
                project_name=start_span_arguments.project_name,
            )
            TRACES_CREATED_BY_DECORATOR.add(trace_data.id)

            span_data = span.SpanData(
                id=helpers.generate_id(),
                parent_span_id=None,
                trace_id=trace_data.id,
                start_time=datetime_helpers.local_timestamp(),
                name=start_span_arguments.name,
                type=start_span_arguments.type,
                tags=start_span_arguments.tags,
                metadata=start_span_arguments.metadata,
                input=start_span_arguments.input,
                project_name=start_span_arguments.project_name,
            )

            context_storage.set_trace_data(trace_data)
            context_storage.add_span_data(span_data)
            return

    def _create_distributed_node_root_span(
        self,
        start_span_arguments: arguments_helpers.StartSpanParameters,
        distributed_trace_headers: DistributedTraceHeadersDict,
    ) -> None:
        span_data = span.SpanData(
            id=helpers.generate_id(),
            parent_span_id=distributed_trace_headers["opik_parent_span_id"],
            trace_id=distributed_trace_headers["opik_trace_id"],
            name=start_span_arguments.name,
            input=start_span_arguments.input,
            metadata=start_span_arguments.metadata,
            tags=start_span_arguments.tags,
            type=start_span_arguments.type,
            project_name=start_span_arguments.project_name,
        )
        context_storage.add_span_data(span_data)

    def _after_call(
        self,
        output: Optional[Any],
        capture_output: bool,
        generators_span_to_end: Optional[span.SpanData] = None,
        generators_trace_to_end: Optional[trace.TraceData] = None,
        flush: bool = False,
    ) -> None:
        try:
            if output is not None:
                end_arguments = self._end_span_inputs_preprocessor(
                    output=output,
                    capture_output=capture_output,
                )
            else:
                end_arguments = arguments_helpers.EndSpanParameters()

            if generators_span_to_end is None:
                span_data_to_end, trace_data_to_end = pop_end_candidates()
            else:
                span_data_to_end, trace_data_to_end = (
                    generators_span_to_end,
                    generators_trace_to_end,
                )

            client = opik_client.get_client_cached()

            span_data_to_end.init_end_time().update(
                **end_arguments.to_kwargs(),
            )

            client.span(**span_data_to_end.__dict__)

            if trace_data_to_end is not None:
                trace_data_to_end.init_end_time().update(
                    **end_arguments.to_kwargs(ignore_keys=["usage"]),
                )

                client.trace(**trace_data_to_end.__dict__)

            if flush:
                client.flush()

        except Exception as exception:
            LOGGER.error(
                logging_messages.UNEXPECTED_EXCEPTION_ON_SPAN_FINALIZATION_FOR_TRACKED_FUNCTION,
                output,
                str(exception),
                exc_info=True,
            )

    def _generators_handler(
        self,
        output: Any,
        capture_output: bool,
        generations_aggregator: Optional[Callable[[List[Any]], str]],
    ) -> Optional[Union[Generator, AsyncGenerator]]:
        """
        Subclasses can override this method to customize generator objects handling
        """
        if inspect.isgenerator(output):
            span_to_end, trace_to_end = pop_end_candidates()
            # For some reason mypy things wrap_sync_generator returns Any
            return generator_wrappers.wrap_sync_generator(  # type: ignore[no-any-return]
                generator=output,
                capture_output=capture_output,
                span_to_end=span_to_end,
                trace_to_end=trace_to_end,
                generations_aggregator=generations_aggregator,
                finally_callback=self._after_call,
            )

        if inspect.isasyncgen(output):
            span_to_end, trace_to_end = pop_end_candidates()
            # For some reason mypy things wrap_async_generator returns Any
            return generator_wrappers.wrap_async_generator(  # type: ignore[no-any-return]
                generator=output,
                capture_output=capture_output,
                span_to_end=span_to_end,
                trace_to_end=trace_to_end,
                generations_aggregator=generations_aggregator,
                finally_callback=self._after_call,
            )

        NOT_A_GENERATOR = None

        return NOT_A_GENERATOR

    @abc.abstractmethod
    def _start_span_inputs_preprocessor(
        self,
        func: Callable,
        name: Optional[str],
        type: SpanType,
        tags: Optional[List[str]],
        metadata: Optional[Dict[str, Any]],
        capture_input: bool,
        args: Tuple,
        kwargs: Dict[str, Any],
        project_name: Optional[str],
    ) -> arguments_helpers.StartSpanParameters: ...

    @abc.abstractmethod
    def _end_span_inputs_preprocessor(
        self,
        output: Optional[Any],
        capture_output: bool,
    ) -> arguments_helpers.EndSpanParameters: ...


def pop_end_candidates() -> Tuple[span.SpanData, Optional[trace.TraceData]]:
    """
    Pops span and trace (if trace exists) data created by @track decorator
    from the current context, returns popped objects.

    Decorator can't attach any child objects to the popped ones because
    they are no longer in context stack.
    """
    span_data_to_end = context_storage.pop_span_data()
    assert (
        span_data_to_end is not None
    ), "When pop_end_candidates is called, top span data must not be None. Otherwise something is wrong."

    trace_data_to_end = None

    possible_trace_data_to_end = context_storage.get_trace_data()
    if (
        context_storage.span_data_stack_empty()
        and possible_trace_data_to_end is not None
        and possible_trace_data_to_end.id in TRACES_CREATED_BY_DECORATOR
    ):
        trace_data_to_end = context_storage.pop_trace_data()
        TRACES_CREATED_BY_DECORATOR.discard(possible_trace_data_to_end.id)

    return span_data_to_end, trace_data_to_end



import logging
from typing import Generator, Any, List, Optional, Callable, Dict
from opik.api_objects import trace, span
from opik.decorator import generator_wrappers

from botocore import eventstream

LOGGER = logging.getLogger(__name__)


def wrap_stream(
    stream: eventstream.EventStream,
    capture_output: bool,
    span_to_end: span.SpanData,
    trace_to_end: Optional[trace.TraceData],
    generations_aggregator: Callable[[List[Any]], Any],
    response_metadata: Dict[str, Any],
    finally_callback: generator_wrappers.FinishGeneratorCallback,
) -> Generator[Any, None, None]:
    items: List[Dict[str, Any]] = []

    try:
        for item in stream:
            items.append(item)

            yield item

    finally:
        aggregated_output = generations_aggregator(items)
        aggregated_output["ResponseMetadata"] = response_metadata

        finally_callback(
            output=aggregated_output,
            generators_span_to_end=span_to_end,
            generators_trace_to_end=trace_to_end,
            capture_output=capture_output,
        )
